optim2by2nonpooled <- function(p,p1,q1,p0,ptd,fdr) {

#
# This function computes the first and second stage critical values and sample sizes of the two-stage 
# design for allele-based test (2X2 tables) with minimal genotyping burden for given PTD 
# and pFDR, when data are NOT POOLED. The calculations are based on the COST-EFFECTIVE APPROXIMATION.
#
# This function uses an algorithm given in the article
#  "J\'{o}zsef Buksz\'{a}r and Edwin J. C. G. van den Oord
#   Optimization of two-stage genetic designs where data are pooled 
#   usingan accurate and efficient approximation for Pearson's statistic,
#   Biometrics, 2005 (submitted)"
#
# Questions about the codes should be addressed to
#  Jo'zsef Buksza'r 
#  email: jbukszar@vcu.edu
#
#
# 
# Input parameters:
#       p  =  proportion of controls in the sample size
#       p1 =  probability that a control falls in category 1 under the alternative hypothesis
#       q1 =  probability that a case falls in category 1 under the alternative hypothesis
#       p0 =  proportion of true null hypotheses
#      ptd =  proportion of true discoveries (PTD) (this can also be interpreted as average power)
#      fdr =  positive false discovery rate (pFDR).
#
#  PACKAGE NEEDED: none
#  FUNCTIONS NEEDED: myoptimalnonnew, refineFNP (directly)
#                    n2minsimpFNP, rminimizernon, critcalNP, condevalnonsimp, Gcritsimp (indirectly)
#
#
#  IN OUTPUT, SAMPLE SIZE IS THE NUMBER OF INDIVIDUALS RECRUITED FOR THE EXPERIMENT !!!                              
#             

   res <- myoptimalnonnew(p,p1,q1,p0,ptd,fdr)

   c10 <- res[2]
   c20 <- res[3]
   n10 <- 2*round(0.5*res[4]+0.25)
   n20 <- 2*round(0.5*res[5]+0.25)


   print( " END RESULT: " )

   refineFNP(c10,c20,n10,n20,p,p1,q1,p0,ptd,fdr)

}



myoptimalnonnew <- function(p,p1,q1,p0,ptd,fdr) {    

# IT RETURNS VERY ACCURATE RESULTS!
# It finds the minimal genotyping burden calculated by "condevalnonsimp"
# for the variables c (first stage critical value) and r (=sqrt(rn/(rn+rnprime)).
#
# PACKAGE NEEDED: -
# FUNCTION NEEDED: condevalnonsimp, rminimizernon, (directly)
#                  Gcritsimp (indirectly)

  q <- 1 - p
  p2 <- 1-p1
  q2 <- 1-q1

  mubase <- (p1-q1)*sqrt( p*q/((p1*p+q1*q)*(p2*p+q2*q)) )

  # Calculating szigma
  aux1 <- ((p1-q1)*q)/(2*(p1*p+q1*q)) + 1
  aux2 <- ((p2-q2)*q)/(2*(p2*p+q2*q)) + 1
  bux1 <- -((p1-q1)*p)/(2*(p1*p+q1*q)) + 1
  bux2 <- -((p2-q2)*p)/(2*(p2*p+q2*q)) + 1

  B11 <- (aux1*aux1*p*q1*q2 + bux1*bux1*q*p1*p2) / (p1*p+q1*q)
  B22 <- (aux2*aux2*p*q1*q2 + bux2*bux2*q*p1*p2) / (p2*p+q2*q)
  B12 <- -(aux1*aux2*p*q1*q2 + bux1*bux2*q*p1*p2) / sqrt( (p1*p+q1*q)*(p2*p+q2*q) )
  B21 <- B12

  bb <- -(B11+B22)
  cc <- B11*B22-B12*B21

  eig1 <- ( -bb + sqrt(bb*bb-4*cc) ) / 2
  eig2 <- ( -bb - sqrt(bb*bb-4*cc) ) / 2

  szigma <- sqrt( max(eig1,eig2) )

  szigmaold <- sqrt( (p1*p2*q+q1*q2*p) / ((p1*p+q1*q)*(p2*p+q2*q)) )

  # End of szigma calculation

  aux <- (ptd*fdr) / (p0*(1-fdr))
  b <- aux * (1-p0)
  aux = qnorm(0.5*b)
  cmax <- aux * aux

# --------------------------

  cold <- 4
  r0 <- 0.2
  cdelta <- 0.00001

  old <- 0
  new <- 0
  cnew <- 0

  cveryold <- 0
  veryold <- 0


  res <- rminimizernon(cold,r0,p0,ptd,fdr,mubase,szigma)   
  old = res[1]
  r0 = res[2]
#  print(res)
  res <- rminimizernon(cold+cdelta,r0,p0,ptd,fdr,mubase,szigma) 
  new = res[1]
#  print(res)

  if (new < old) cstep <- 0.5 else cstep <- -0.5
  cnew = cold + cstep
  res <- rminimizernon(cnew,r0,p0,ptd,fdr,mubase,szigma) 
  new = res[1]
#  print(res)

  cveryold = cold
  veryold = old
  while (new < old) {
      cveryold = cold
      veryold = old
      cold = cnew
       old = new
      cnew = cold + cstep
      res <- rminimizernon(cnew,r0,p0,ptd,fdr,mubase,szigma) 
      new = res[1]
      r0 = res[2]
#      print(res)
  }

  if (cveryold < cnew) {
     a <- cveryold
     b <- cnew
     fva <- veryold
     fvb <- new
  } else {
     b <- cveryold
     a <- cnew
     fvb <- veryold
     fva <- new
  }

  c <- (a+b)*0.5
  res <- rminimizernon(c,r0,p0,ptd,fdr,mubase,szigma) 
  fvc <- res[1]

  iter <- 0
  while ( b-a > 0.00001 ) {
     iter = iter + 1

     if ( fvc > fva ) {
        b = c
        fvb = fvc
        c = (a+b)*0.5
        res <- rminimizernon(c,r0,p0,ptd,fdr,mubase,szigma)
        fvc = res[1]
        r0 = res[2]
     } else {
         if ( fvc > fvb ) {
            a = c
            fva = fvc
            c = (a+b)*0.5
            res <- rminimizernon(c,r0,p0,ptd,fdr,mubase,szigma)
            fvc = res[1]
            r0 = res[2]
         } else {
            a = (a+c)*0.5
            b = (b+c)*0.5
            res <- rminimizernon(a,r0,p0,ptd,fdr,mubase,szigma)
            fva = res[1]
            res <- rminimizernon(b,r0,p0,ptd,fdr,mubase,szigma)
            fvb = res[1]
            r0 = res[2]
         }
     }    

  } # end while

  if ( fvc > fva ) {
     crit1 <- a
     gb <- fva
  } else {
      if ( fvc > fvb ) {
          crit1 <- b
          gb <- fvb
      } else {
          crit1 <- c
          gb <- fvc
      }
  }    

  resul <- rminimizernon(crit1,r0,p0,ptd,fdr,mubase,szigma)
  rold <- resul[2]
  res <- condevalnonsimp(crit1,rold,p0,ptd,fdr,mubase,szigma)

  res

}



refineFNP <- function(c10,c20,n10,n20,p,p1,q1,p0,ptd,fdr) {

#  FOR ALLELE BASED TESTS ON NON-POOLED DATA
#  ( SAMPLE SIZE IS THE NUMBER OF INDIVIDUALS RECRUITED FOR THE EXPERIMENT )
#  BASED ON COSTEFFECTIVE APPROXIMATION
#
#  PACKAGE NEEDED: -
#  FUNCTION NEEDED: n2minsimpFNP, critcalNP
#                  

  q <- 1 - p
  p2 <- 1-p1
  q2 <- 1-q1

  mubase <- (p1-q1)*sqrt( p*q/((p1*p+q1*q)*(p2*p+q2*q)) )

  # Calculating szigma
  aux1 <- ((p1-q1)*q)/(2*(p1*p+q1*q)) + 1
  aux2 <- ((p2-q2)*q)/(2*(p2*p+q2*q)) + 1
  bux1 <- -((p1-q1)*p)/(2*(p1*p+q1*q)) + 1
  bux2 <- -((p2-q2)*p)/(2*(p2*p+q2*q)) + 1

  B11 <- (aux1*aux1*p*q1*q2 + bux1*bux1*q*p1*p2) / (p1*p+q1*q)
  B22 <- (aux2*aux2*p*q1*q2 + bux2*bux2*q*p1*p2) / (p2*p+q2*q)
  B12 <- -(aux1*aux2*p*q1*q2 + bux1*bux2*q*p1*p2) / sqrt( (p1*p+q1*q)*(p2*p+q2*q) )
  B21 <- B12

  bb <- -(B11+B22)
  cc <- B11*B22-B12*B21

  eig1 <- ( -bb + sqrt(bb*bb-4*cc) ) / 2
  eig2 <- ( -bb - sqrt(bb*bb-4*cc) ) / 2

  szigma <- sqrt( max(eig1,eig2) )

  szigmaold <- sqrt( (p1*p2*q+q1*q2*p) / ((p1*p+q1*q)*(p2*p+q2*q)) )

  # End of szigma calculation


  aux <- (ptd*fdr) / (p0*(1-fdr))
  b <- aux * (1-p0)

  old <- 0
  new <- 0
  rnew <- 0

  rold <- n10
  rveryold <- 0
  veryold <- 0


  res <- n2minsimpFNP(c10,c20,rold,n20,mubase,szigma,p0,ptd,fdr)
  old = res[1]

  res <- n2minsimpFNP(c10,c20,rold+1,res[2],mubase,szigma,p0,ptd,fdr)
  new = res[1]

  if (new < old) rstep <- 4 else rstep <- -4
  rnew = rold + rstep
  res <- n2minsimpFNP(c10,c20,rnew,res[2],mubase,szigma,p0,ptd,fdr)
  new = res[1]


  rveryold = rold
  veryold = old

  while (new < old) {
      rveryold = rold
      veryold = old
      rold = rnew
       old = new
      rnew = rold + rstep
      res <- n2minsimpFNP(c10,c20,rnew,res[2],mubase,szigma,p0,ptd,fdr)
      new = res[1]

  }

  if (rveryold < rnew) {
     a <- rveryold
     b <- rnew
     fva <- veryold
     fvb <- new
  } else {
     b <- rveryold
     a <- rnew
     fvb <- veryold
     fva <- new
  }

  c <- (a+b)*0.5
  res <- n2minsimpFNP(c10,c20,c,res[2],mubase,szigma,p0,ptd,fdr)
  fvc <- res[1]

  iter <- 0
  while ( b-a > 4 ) {
     iter = iter + 1
     if ( fvc > fva ) {
        b = c
        fvb = fvc
        c = (a+b)*0.5
        res <- n2minsimpFNP(c10,c20,c,res[2],mubase,szigma,p0,ptd,fdr)
        fvc = res[1]
     } else {
         if ( fvc > fvb ) {
            a = c
            fva = fvc
            c = (a+b)*0.5
            res <- n2minsimpFNP(c10,c20,c,res[2],mubase,szigma,p0,ptd,fdr)
            fvc = res[1]
         } else {
            a = (a+c)*0.5
            b = (b+c)*0.5
            res <- n2minsimpFNP(c10,c20,a,res[2],mubase,szigma,p0,ptd,fdr)
            fva = res[1]
            res <- n2minsimpFNP(c10,c20,b,res[2],mubase,szigma,p0,ptd,fdr)
            fvb = res[1]
         }
     } 


  } # end while

  if ( fvc > fva ) {
     value <- c(fva,a)
  } else {
      if ( fvc > fvb ) {
          value <- c(fvb,b)
      } else {
          value <- c(fvc,c)
      }
  }    

  genburd <- value[1]
  n1 <- value[2]
  res <- n2minsimpFNP(c10,c20,n1,res[2],mubase,szigma,p0,ptd,fdr)
  n2 = res[2]
  res <- critcalNP(c10,c20,n1,n2,mubase,szigma,p0,ptd,fdr)
  c1 <- res[1]
  c2 <- res[2]

  print(" First and second stage critical value = ")
  print( c(c1, c2) )

  print(" First and second stage sample size = ")
  print( c(n1/2, n2/2) )

  print(" genotyping burden = ")
  print( genburd/2 )
 
}



n2minsimpFNP <- function(c10,c20,n1,n20,mubase,szigma,p0,ptd,fdr) {

# 
#  FOR ALLELE BASED TESTS (HALF SAMPLE SIZE FOR 2 X 2 TABLES)
#  BASED ON COSTEFFECTIVE METHOD
#
#  PACKAGE NEEDED: -
#  FUNCTION NEEDED: critcalNP
#                  

  old <- 0
  new <- 0
  rnew <- 0

  rold <- n20
  rveryold <- 0
  veryold <- 0

  res <- critcalNP(c10,c20,n1,rold,mubase,szigma,p0,ptd,fdr)
  old = res[3]

  res <- critcalNP(c10,c20,n1,rold+1,mubase,szigma,p0,ptd,fdr)
  new = res[3]

  if (new < old) rstep <- 8 else rstep <- -8
  rnew = rold + rstep
  res <- critcalNP(c10,c20,n1,rnew,mubase,szigma,p0,ptd,fdr)
  new = res[3]


  rveryold = rold
  veryold = old
  while (new < old) {
      rveryold = rold
      veryold = old
      rold = rnew
       old = new
      rnew = rold + rstep
      res <- critcalNP(c10,c20,n1,rnew,mubase,szigma,p0,ptd,fdr)
      new = res[3]

  }

  if (rveryold < rnew) {
     a <- rveryold
     b <- rnew
     fva <- veryold
     fvb <- new
  } else {
     b <- rveryold
     a <- rnew
     fvb <- veryold
     fva <- new
  }

  c <- (a+b)*0.5
  resc <- critcalNP(c10,c20,n1,c,mubase,szigma,p0,ptd,fdr)
  fvc <- resc[3]

  iter <- 0
  while ( b-a > 4 ) {
     iter = iter + 1
     if ( fvc > fva ) {
        b = c
        fvb = fvc
        c = (a+b)*0.5
        resc <- critcalNP(c10,c20,n1,c,mubase,szigma,p0,ptd,fdr)
        fvc = resc[3]
     } else {
         if ( fvc > fvb ) {
            a = c
            fva = fvc
            c = (a+b)*0.5
            resc <- critcalNP(c10,c20,n1,c,mubase,szigma,p0,ptd,fdr)
            fvc = resc[3]
         } else {
            a = (a+c)*0.5
            b = (b+c)*0.5
            resa <- critcalNP(c10,c20,n1,a,mubase,szigma,p0,ptd,fdr)
            fva = resa[3]
            resb <- critcalNP(c10,c20,n1,b,mubase,szigma,p0,ptd,fdr)
            fvb = resb[3]
         }
     } 

  } # end while

  if ( fvc > fva ) {
     value <- c(fva,a)
  } else {
      if ( fvc > fvb ) {
          value <- c(fvb,b)
      } else {
          value <- c(fvc,c)
      }
  }    
 
  value

}



rminimizernon <- function(cold,r0,p0,ptd,fdr,mubase,szigma) {    

# For a given first stage critical value, cold, it calculates r
# that minimizes the genotyping burden in the unconstrained NLP.
#
# PACKAGE NEEDED: -
# FUNCTION NEEDED: condevalnonsimp
#                  Gcritsimp (indirectly)


  rdelta <- 0.000001

  old <- 0
  new <- 0
  rnew <- 0

  rold <- r0
  rveryold <- 0
  veryold <- 0

  res <- condevalnonsimp(cold,rold,p0,ptd,fdr,mubase,szigma)
  old = res[1]
  res <- condevalnonsimp(cold,rold+rdelta,p0,ptd,fdr,mubase,szigma)
  new = res[1]

  if (new < old) rstep <- 0.1 else rstep <- -0.1
  rnew = rold + rstep
  res <- condevalnonsimp(cold,rnew,p0,ptd,fdr,mubase,szigma)
  new = res[1]

  rveryold = rold
  veryold = old
  while (new < old) {
      rveryold = rold
      veryold = old
      rold = rnew
       old = new
      rnew = rold + rstep
      res <- condevalnonsimp(cold,rnew,p0,ptd,fdr,mubase,szigma)
      new = res[1]
  }

  if (rveryold < rnew) {
     a <- rveryold
     b <- rnew
     fva <- veryold
     fvb <- new
  } else {
     b <- rveryold
     a <- rnew
     fvb <- veryold
     fva <- new
  }

  c <- (a+b)*0.5
  res <- condevalnonsimp(cold,c,p0,ptd,fdr,mubase,szigma)
  fvc <- res[1]

  iter <- 0
  while ( b-a > 0.00001 ) {
     iter = iter + 1
     if ( fvc > fva ) {
        b = c
        fvb = fvc
        c = (a+b)*0.5
        res <- condevalnonsimp(cold,c,p0,ptd,fdr,mubase,szigma)
        fvc = res[1]
     } else {
         if ( fvc > fvb ) {
            a = c
            fva = fvc
            c = (a+b)*0.5
            res <- condevalnonsimp(cold,c,p0,ptd,fdr,mubase,szigma)
            fvc = res[1]
         } else {
            a = (a+c)*0.5
            b = (b+c)*0.5
            res <- condevalnonsimp(cold,a,p0,ptd,fdr,mubase,szigma)
            fva = res[1]
            res <- condevalnonsimp(cold,b,p0,ptd,fdr,mubase,szigma)
            fvb = res[1]
         }
     }    

  } # end while

  if ( fvc > fva ) {
     value <- c(fva,a)
  } else {
      if ( fvc > fvb ) {
          value <- c(fvb,b)
      } else {
          value <- c(fvc,c)
      }
  }    

  value

}



critcalNP <- function(c10,c20,n1,n2,mubase,szigma,p0,ptd,fdr) {

#  FOR 2 X 2 TABLES
#  BASED ON COSTEFFECTIVE APPROXIMATION
#  Calculates the first and second stage critical values for the 
#  first and second stage sample sizes so that both constraints hold. 
#
#  PACKAGE NEEDED: -
#  FUNCTION NEEDED: Gcritsimp
#                  

    r <- sqrt(n1/(n1+n2))

    a <- ptd
    aux <- (ptd*fdr) / (p0*(1-fdr))
    b <- aux * (1-p0)

    preject1 <- function(c1x) { 
        c2x <- qchisq(1 - ( b / (1-pchisq(c1x,1)) ),1)
        value <- Gcritsimp(c1x,n1,mubase,szigma) * Gcritsimp(c2x,n2,mubase,szigma)  - a
        value
    }

    top <- qchisq(1-b,1) - 0.1
    lo <- preject1(1)
    up <- preject1(top)
    if (lo*up>0 || is.nan(lo*up)) {
        print(" PROBLEM FOUND IN critcalNP: UNIROOT CANNOT BE USED!!!!!")
        print(c(lo,up))
    } else {
      o   <- uniroot(preject1, lower = 1, upper = top, tol = 0.0000001)
      c1 <- o$root[1]  
      c2 <- qchisq(1 - ( b / (1-pchisq(c1,1)) ),1)  

      genburd <- n1 + n2*( p0*(1-pchisq(c1,1)) +(1-p0)*Gcritsimp(c1,n1,mubase,szigma) ) 
    }        

    value <- c(c1,c2,genburd)

    value
    
}



condevalnonsimp <- function(c,r,p0,ptd,fdr,mubase,szigma) {    

# ASSUMES NONPOOLED DATA
# For a first stage critical value and r=sqrt(rn/rntot), it calculates 
# the second stage critical value and sample size and the genotyping burden
# in such a way that the two conditions in the NLP hold. 
#
# PACKAGE NEEDED: -
# FUNCTION NEEDED:  Gcritsimp

    gamma <- mubase

    a <- ptd
    aux <- (ptd*fdr) / (p0*(1-fdr))
    b <- aux * (1-p0)
    rcoeff <- 1/(r*r) - 1

    aux <- qnorm(  b/( 4*pnorm(-sqrt(c)) )  )
    cs <- aux * aux

    mrootcs <- -sqrt(cs) / szigma
    mrootc <- -sqrt(c) / szigma
    gamsz <- gamma / szigma
    gamperr <- gamsz * sqrt(rcoeff)

    preject1 <- function(nx) { 
        rootn <- sqrt(nx)
        value <- ( pnorm(mrootcs+gamperr*rootn) + pnorm(mrootcs-gamperr*rootn) ) * ( pnorm(mrootc+gamsz*rootn) + pnorm(mrootc-gamsz*rootn) )  - a
        value
    }
    lo <- preject1(1)
    up <- preject1(100000)
    if (lo*up>0 || is.nan(lo*up)) {
        print(c(lo,up))
        print(c(c,r,cs))
        genburd <- 10000000
    } else {
      o   <- uniroot(preject1, lower = 1, upper = 100000, tol = 0.00001)
      rn <- o$root[1]  # The second condition holds with this rn (first stage sample size) and 
      rnprime <- rcoeff * rn           # this rnprime (second stage sample size).

      genburd <- rn + rnprime*( p0*(1-pchisq(c,1)) +(1-p0)*Gcritsimp(c,rn,mubase,szigma) ) 

    }
                                           
    value <- c(genburd,c,cs,rn,rnprime)

    value

}



Gcritsimp <- function(c,rn,mubase,szigma) {    

# BASED ON THE COST-EFFECTIVE APPROXIMATION!
# Calculates the probability of rejecting a marker in a single stage when the critical value is c,
# the sample size is rn under H_1, i.e. when the probability table is given by p1 and q1.
# Number of controls is p*rn, number of cases is q*rn.
#
# FUNCTION NEEDED: -
     
      coeffo <- 1 / (szigma*szigma)

      nonco <- coeffo * mubase * mubase * rn
      value <- 1 - pchisq(coeffo*c, 1, ncp=nonco)

      value
}
